/* sapf.h
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   header file for the Simple Active Packet Format
 *
 * Copyright (C) 2001-2004 <christian.tschudin@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 */

#ifndef _SAPF_H
#define _SAPF_H

typedef __u64 sapfsel_t; 

typedef struct sapfpkt_s {
	int len;	// len of data
	short ifnr;	// incoming interface number
	short snr;	// SNR of received packet (not used yet)
	char *data;	// raw, not including selector
} sapfpkt;

typedef void (NS_CLASS *sapfcallback_t)(sapfpkt *p, void *aux);

//COMMON DEFINITIONS

#define SELCLASS_MASK   const64u(0x0c00000000000)
#define SELCLASS_STD    const64u(0x0000000000000)
#define SELCLASS_LOCAL  const64u(0x0400000000000) 
#define SELCLASS_REMOTE const64u(0x0800000000000) 

#define selIsStd(s)    ((ntohsel(s) & SELCLASS_MASK) == SELCLASS_STD)
#define selIsLocal(s)  ((ntohsel(s) & SELCLASS_MASK) == SELCLASS_LOCAL)
#define selIsRemote(s) ((ntohsel(s) & SELCLASS_MASK) == SELCLASS_REMOTE)

#define SAPF_ETHERTYPE 0x4242

#define SAPF_SELNET_XRP            const64u(0x0000000000000001)
#define SAPF_SELNET_LUNAR_RREQ     const64u(0x0000000000000002)
#define SAPF_SELNET_LUNAR_BCAST    const64u(0x0000000000000003)


// ----------------------------------------------------------------------
// private data structures:

struct fibentry_s {
	unsigned char client;
	unsigned char flags_and_op;

#define FE_MASK      0x03
#define FE_EMPTY     0x00
#define FE_USED      0x01
#define FE_DELETED   0x02

#define SAPFOP_SHIFT 2
#define SAPFOP_MASK  (0x03 << SAPFOP_SHIFT)
#define SAPFOP_CBACK (0x01 << SAPFOP_SHIFT)
#define SAPFOP_REDIR (0x02 << SAPFOP_SHIFT)

#define FE_ACTIVE    0x10

    short credit;        // -1 means: permanent
    sapfsel_t sel;
    union {
	sapfsel_t sel;
	sapfcallback_t callback;
    } p1;
    union {
	sapfsel_t sel;
	void *aux;
    } p2;
};

struct fibtable_s {
	sapfsel_t nextselval;
	char name[16];
	int fibsize;
	int opencnt;
	struct fibentry_s tab[0];
};

#define SELMASK const64u(0x0000ffffffffffff)

#define FIB_START_CREDIT  2048
#define FIB_DECAY_PER_SEC  128
#define FIB_DECAY_PER_PKT    1


// ----------------------------------------------------------------------
// implementation limits:

#define SAPF_MAX_FIB_TABLES 5
#define SAPF_MAX_CLIENTS 32

#endif /* _SAPF_H */
