# LUNAR.tcl

set val(chan)       Channel/WirelessChannel
set val(prop)       Propagation/TwoRayGround
set val(netif)      Phy/WirelessPhy
set val(mac)        Mac/802_11
set val(ifq)        Queue/DropTail/PriQueue
set val(ll)         LL
set val(ant)        Antenna/OmniAntenna
set val(x)              1200
set val(y)              1200
set val(ifqlen)         50            ;# max packet in ifq
set val(seed)           0.82
set val(adhocRouting)   LUNAR
set val(nn)             4     ;# how many nodes are simulated
set val(cp)             "test.cpattern" 
set val(sc)             "test.scenario" 
set val(stop)           50           ;# simulation time
set val(nnhops)          3           ;# simulation time
set val(beams)           1           ;# simulation time

add-packet-header SELNET XRP

#set up the antennas to be centered in the node and 1.5 meters above it
Antenna/OmniAntenna set X_ 0
Antenna/OmniAntenna set Y_ 0
Antenna/OmniAntenna set Z_ 1.5
Antenna/OmniAntenna set Gt_ 1.0
Antenna/OmniAntenna set Gr_ 1.0

# Initialize the SharedMedia interface with parameters to make
# it work like the 914MHz Lucent WaveLAN DSSS radio interface
Phy/WirelessPhy set CPThresh_ 10.0
Phy/WirelessPhy set CSThresh_ 1.559e-11
Phy/WirelessPhy set RXThresh_ 3.652e-10
Phy/WirelessPhy set Rb_ 2e6
Phy/WirelessPhy set Pt_ 0.2818
Phy/WirelessPhy set freq_ 914e+6
Phy/WirelessPhy set L_ 1.0
 
Mac/802_11 set dataRate_ 2e6

set ns_		[new Simulator]


set topo	[new Topography]



set tracefd	[open lunar_test.tr w]
#set namtrace    [open LUNAR.nam w]


$ns_ trace-all $tracefd 
#$ns_ namtrace-all-wireless $namtrace $val(x) $val(y)

$topo load_flatgrid $val(x) $val(y)

set god_ [create-god $val(nn)]

# Create channel
set chan_1_ [new $val(chan)]

$ns_ node-config -adhocRouting $val(adhocRouting) \
                 -llType $val(ll) \
                 -macType $val(mac) \
                 -ifqType $val(ifq) \
                 -ifqLen $val(ifqlen) \
                 -antType $val(ant) \
                 -propType $val(prop) \
                 -phyType $val(netif) \
		 -topoInstance $topo \
		 -agentTrace ON \
                 -routerTrace OFF \
                 -macTrace ON \
		 -movementTrace ON \
		 -channel $chan_1_


SELNET set conf_max_ring                3

Queue/DropTail/PriQueue set Prefer_Routing_Protocols    1

for {set i 1} {$i <= $val(nn) } {incr i} {
	set node_($i) [$ns_ node]	
	$node_($i) random-motion 0	;

        if {$val(adhocRouting) != "LUNAR"} {

            set r [$node_($i) set ragent_]
                $r set debug_ 1
                $r set rt_log_interval_ 1000
                $r set log_to_file_ 1
                $r set local_repair_ 1
                $r set llfeedback_ 0
                $r set hello_jittering_ 1
                $r set rreq_gratuitous_ 0
                $r set wait_on_reboot_ 0
            }
    }
puts "Loading connection pattern..."
source $val(cp)

puts "Loading scenario file..."
source $val(sc)

for {set i 1} {$i <= $val(nn)} {incr i} {

    $ns_ initial_node_pos $node_($i) 20
}


for {set i 1} {$i <= $val(nn) } {incr i} {
    $ns_ at $val(stop).0 "$node_($i) reset";
}

$ns_ at  $val(stop).0002 "puts \"NS EXITING...\" ; $ns_ halt"

puts $tracefd "M 0.0 nn $val(nn) x $val(x) y $val(y) rp $val(adhocRouting)"
puts $tracefd "M 0.0 sc $val(sc) cp $val(cp) seed $val(seed)"
puts $tracefd "M 0.0 prop $val(prop) ant $val(ant)"

puts "Starting Simulation..."
$ns_ run

