/* lunar.c
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   Lunar logic
 *
 * Copyright (C) 2001-2003 <christian.tschudin@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 */


#include "../include/lunar.h"
#include "../include/lunar_decl.h"

// patch-me: we should pass "struct sockaddr" instead of a 32bit integer
// for the src parameter, so this cludge below would go away:
#ifndef NS_PORT
# define NO_IP_ADDR 0
#else
# define NO_IP_ADDR -1
#endif

// ----------------------------------------------------------------------

void NS_CLASS lunar_bcast_handler(sapfpkt *p, void *aux)
{
}

void NS_CLASS schedule_rreq(void *data, int len, unsigned long usec)
{
	int ifnr = basedev ? basedev->ifindex : 0;

	NETBOX_Instance netbox_l2send(ifnr, 0,
				      htonsel(SAPF_SELNET_LUNAR_RREQ),
				      data, len);
}

 int NS_CLASS peer_is_unknown(u_int32_t name)
{
	int i;

	for (i = 0; i < LUNAR_MAX_PEERS; i++)
		if (peers[i].used && peers[i].ia.s_addr == name)
			return 0;
	return 1;
}

 int NS_CLASS add_peer(struct in_addr name)
{
	int i, ndx = -1;

	for (i = 0; i < LUNAR_MAX_PEERS; i++) {
		if (ndx < 0 && !peers[i].used)
			ndx = i;
		else if (peers[i].ia.s_addr == name.s_addr)
			return i;
	}
	if (ndx >= 0) {

#ifndef NS_PORT
	  memset(peers + ndx, 0, sizeof(struct peer_s));
#endif
		peers[ndx].ia = name;
		peers[ndx].hops = 0;
	}
	return ndx;
}

// ----------------------------------------------------------------------

/*   mkSapfDest
 *   allocates an eth+selector (netorder) buffer, to be used as a
 *   network pointer in the SAPF module. The SAPF module will free
 *   this netptr when the SAPF entry is removed.
 */
struct sapfdest_s* NS_CLASS mkSapfDest(unsigned char *eth, sapfsel_t sel)
{
	struct sapfdest_s *dst =(struct sapfdest_s*)  kmalloc(sizeof(struct sapfdest_s), GFP_ATOMIC);
	unsigned char *cp;

	if (!dst)
		return 0;

	cp = eth ? eth : (unsigned char *) "\xff\xff\xff\xff\xff\xff";

	DEBUG(10, printk("mkSapfDest %02x:%02x:%02x:%02x:%02x:%02x 0x%016Lx\n",
		 cp[0], cp[1], cp[2], cp[3], cp[4], cp[5], ntohsel(sel)));

	dst->ifnr = basedev ? basedev->ifindex : 0;
	dst->sel = sel;
	memcpy(dst->dst.sa_data, cp, ETH_ALEN);

	return dst;
}

/*   mkNetPointer
 *   installs a forwarder entry, returns the new selector.
 */
sapfsel_t NS_CLASS mkNetPointer(unsigned char *eth, sapfsel_t sel)
{
	sapfsel_t ptr = SAPF_Instance sapf_getUnusedSelector(mysapfhandle, SELCLASS_LOCAL);
        SAPF_Instance sapf_addCallback(mysapfhandle, ptr,NS_CLASS_addr L2sendto, mkSapfDest(eth, sel));
	return ptr;
}

sapfsel_t NS_CLASS mkHandlerPointer(sapfcallback_t fct, void *aux)
{
	sapfsel_t ptr =SAPF_Instance sapf_getUnusedSelector(mysapfhandle, SELCLASS_LOCAL);
        SAPF_Instance sapf_addCallback(mysapfhandle, ptr, fct, aux);
	return ptr;
}


// ----------------------------------------------------------------------

 void NS_CLASS arp_answer(int ndx)
{
	struct sockaddr sa;
	// sa.sa_family = ARPHRD_ETHER;
	// we use the low 48 bits, and all selectors are stored in nw order:
	memcpy(sa.sa_data,
	       ((char*) &(peers[ndx].fwdsel)) + sizeof(sapfsel_t) - ETH_ALEN, ETH_ALEN);
	NETBOX_Instance netbox_arpset(peers[ndx].ia, &sa, lunardev);
}


 int NS_CLASS install_peer(struct in_addr ia, sapfsel_t ptr,
			unsigned char plen, char timeroffs)
{
	int ndx;
	unsigned long t;

	DEBUG(10, printk("install peer %08x\n", ntohl(ia.s_addr)));

	ndx = add_peer(ia);
	if (ndx < 0) {
		DEBUG(10, printk("install peer: could not add peer\n"));
		return -1;
	}
	if (peers[ndx].used == -1) { // must be in "resolving state"
		DEBUG(10, printk("install peer: not in resolving state\n"));
		return -1;
	}

	peers[ndx].used = -1;     // now resolved
	peers[ndx].hops = plen;
	peers[ndx].flags &= ~PEER_BIDIR;
	peers[ndx].fwdsel = ptr;

	t = (timeroffs ? conf_fri/2 : conf_fri) -
		((int) SAPF_Instance getRandomByte()-128) *
		conf_fri_jitter/128;

	NETBOX_Instance netbox_deltimer(&(peers[ndx].timer));
	NETBOX_Instance netbox_newtimer(&(peers[ndx].timer), _VOID_VOID_ peer_timeout,
					peers + ndx, t);

	// if (peers[ndx].flags & PEER_ARP_PENDING) {
	//	peers[ndx].flags &= ~PEER_ARP_PENDING;
	if (peers[ndx].ia.s_addr != htonl(LUNAR_IP_BCAST))
		arp_answer(ndx);
	// }

	return 0;
}
// ----------------------------------------------------------------------
//
 void NS_CLASS unicast_rrep_handler(sapfpkt *p, void *ups_)
{
  struct uc_pending_s *ups = (struct uc_pending_s *)ups_;
	struct xrp_head_s *x = (struct xrp_head_s*) p->data;
	struct xrp_sel_eth_s *f;

        DEBUG(9, printk("unicast_rrep_handler (%d bytes)\n", p->len));

	if (p->len < sizeof(struct xrp_head_s) ||
	    x->version != XRP_VERSION_REPLY) {
		DEBUG(9, printk("  unknown or garbled RREP\n"));
		return ;
	}
	f = (struct xrp_sel_eth_s *) xrp_locate(x+1, p->len - sizeof(*x), XRP_CLASS_FWD_ADDR);
	if (!f || f->partype != XRP_PARTYPE_SEL_ETH) {
		DEBUG(9, printk("  no forward addr found in RREP %p\n", f));
		return ;
	}

	if (install_peer(ups->ia, mkNetPointer(f->eth, f->sel), x->ttl, 0))
		return ;

	if (ups->is_bidir)
		SAPF_Instance sapf_addREDIR(mysapfhandle, ups->backsel, ipv4deliver);

        SAPF_Instance sapf_deleteEntry(mysapfhandle, ups->caller);

        return ;
}

 void NS_CLASS unicast_transit_rrep_handler(sapfpkt *p,void *tr_)
{
  struct uc_transit_pending_s *tr =(struct uc_transit_pending_s *)tr_;
	int len, ifnr = basedev ? basedev->ifindex : 0;
	struct xrp_head_s *x;
	struct xrp_sel_eth_s *f;
	struct sockaddr sa;
	char buf[300];

        DEBUG(9, printk("unicast transit rrep handler (%d bytes)\n",
			p->len));

	if (p->len < sizeof(struct xrp_head_s))
		return ;

	len = MIN(sizeof(buf), p->len);
	memcpy(buf, p->data, len);
	x = (struct xrp_head_s *) buf;
	if (x->version != XRP_VERSION_REPLY) {
		DEBUG(9, printk("  unknown or garbled RREP\n"));
		return ;
	}
	x->ttl++;

	f = (struct xrp_sel_eth_s *)
		xrp_locate(x+1, len - sizeof(*x), XRP_CLASS_FWD_ADDR);
	if (!f || f->partype != XRP_PARTYPE_SEL_ETH) {
		DEBUG(9, printk("  no forward addr found in RREP %p\n", f));
		return ;
	}

	if (tr->is_bidir)
		SAPF_Instance sapf_addCallback(mysapfhandle, tr->mybacksel, NS_CLASS_addr L2sendto,
				 mkSapfDest(tr->backeth, tr->backsel));

	NETBOX_Instance netbox_getL2addr(ifnr, &sa);
	f->sel = mkNetPointer(f->eth, f->sel);
	memcpy(f->eth, sa.sa_data, ETH_ALEN);

	memcpy(sa.sa_data, tr->replyeth, ETH_ALEN);
	NETBOX_Instance netbox_l2send(ifnr, &sa, tr->replysel, buf, len);

        SAPF_Instance sapf_deleteEntry(mysapfhandle, tr->caller);

        return ;
}

 void NS_CLASS bcast_rep_handler(sapfpkt *p, void *bps_)
{
  struct bcast_pending_s *bps=(struct bcast_pending_s *)bps_;
	struct xrp_head_s *x = (struct xrp_head_s*) p->data;
	struct xrp_sel_eth_s *f;
	NS_STATIC unsigned char broadcast[6] = {'\xff', '\xff', '\xff', '\xff', '\xff', '\xff'};

        DEBUG(9, printk("bcast_rep_handler (%d bytes)\n", p->len));

	if (p->len < sizeof(struct xrp_head_s) ||
	    x->version != XRP_VERSION_REPLY) {
		DEBUG(9, printk("  unknown or garbled REP\n"));
		return ;
	}
	f = (struct xrp_sel_eth_s *)xrp_locate(x+1, p->len - sizeof(*x), XRP_CLASS_FWD_ADDR);
	if (!f || f->partype != XRP_PARTYPE_SEL_ETH) {
		DEBUG(9, printk("  no forward addr found in RREP %p\n", f));
		return ;
	}
	if (f->sel != bps->fwd_sel) {
		DEBUG(9, printk("  returned bc sel does not match\n"));
		return ;
	}

	bc_sel = mkNetPointer(broadcast, f->sel);
	if (!bc_len) {
		install_peer(bps->ia, bc_sel, 3, 0);
		if (bc_selp)
			*bc_selp = bc_sel;
	}

        SAPF_Instance sapf_deleteEntry(mysapfhandle, bps->caller);

        return ;
}

// ----------------------------------------------------------------------

void NS_CLASS lunar_rreq_handler(sapfpkt *p, void *aux)
{
	int ifnr = basedev ? basedev->ifindex : 0;
	struct xrp_head_s *x = (struct xrp_head_s *) p->data;
	struct xrp_sel_eth_s *reply = 0, *backwd = 0;
	struct xrp_sel_s *series = 0, *fwd = 0;
	struct xrp_ipv4_s *src = 0, *target = 0;
	struct in_addr myip;
	struct sockaddr sa;
	xrp_param_t *pp; // pointer to a single parameter block
	int cnt, len;

	DEBUG(9, printk("LUNAR: RREQ (%d bytes)\n", p->len));

	if (p->len < sizeof(struct xrp_head_s) ||
	    x->version != XRP_VERSION_REQUEST) {
		DEBUG(9, printk(" wrong length or version\n"));
		return;
	}

	len = p->len - sizeof(*x);
	pp = (xrp_param_t *) (x+1);
	while (len >= 4) {
		if (pp->param.len == 0)
			break;
		if (ntohs(pp->param.len) > len ||
		    ntohs(pp->param.len) < 4) {
			DEBUG(9, printk(" end of XRP parsing: wrong length\n"));
			return;
		}
		if (pp->param.partype > XRP_PARTYPE_UNKNOWN &&
		    pp->param.partype < XRP_PARTYPE_COUNT) {
			if (ntohs(pp->param.len) !=
			    xrp_param_len[pp->param.partype]) {
				DEBUG(9, printk(" wrong parameter length, class=%d\n", pp->param.parclass));
				return;
			}
			DEBUG(10, printk(" param class+len+expect_len: %d %d %d\n",
			pp->param.parclass, ntohs(pp->param.len), xrp_param_len[pp->param.partype]));

			switch (pp->param.parclass) {
			case XRP_CLASS_REQ_SERIES:
				if (pp->param.partype == XRP_PARTYPE_SEL)
					series = (struct xrp_sel_s *) pp;
				break;
			case XRP_CLASS_REPLY_ADDR:
				if (pp->param.partype == XRP_PARTYPE_SEL_ETH)
					reply = (struct xrp_sel_eth_s *) pp;
				break;
			case XRP_CLASS_TARGET:
				if (pp->param.partype == XRP_PARTYPE_IPv4)
					target = (struct xrp_ipv4_s*) pp;
				break;
			case XRP_CLASS_BACKWD_ADDR:
				if (pp->param.partype == XRP_PARTYPE_SEL_ETH)
					backwd = (struct xrp_sel_eth_s *) pp;
				break;
			case XRP_CLASS_FWD_ADDR: // for broadcast
				if (pp->param.partype == XRP_PARTYPE_SEL)
					fwd = (struct xrp_sel_s *) pp;
				break;
			case XRP_CLASS_SRC_ADDR:
				if (pp->param.partype == XRP_PARTYPE_IPv4)
					src = (struct xrp_ipv4_s *) pp;
				break;
			default: // ignore unknown fields
				break;
			}
		}
		cnt = (ntohs(pp->param.len) + 3) & ~0x3;
		len -= cnt;
		pp = (xrp_param_t*) (((char*)pp) + cnt);
	}

	if (!series || SAPF_Instance sapf_exists(mysapfhandle, series->sel)) {
		DEBUG(9, printk(" no series/series %016Lx already exists, dropping\n", series ? ntohsel(series->sel) : const64(0)));
		return;
	}

	SAPF_Instance sapf_addREDIR(mysapfhandle, series->sel, 0);

	NETBOX_Instance netbox_getL3addr(&myip, lunardev);
	NETBOX_Instance netbox_getL2addr(ifnr, &sa);

	if (!target || !reply) {
		DEBUG(9, printk(" no target or no reply sel: %p %p\n",
				target, reply));
		return;
	}
	DEBUG(9, printk("  target is IP 0x%08x, I'm 0x%08x\n", ntohl(target->ip), ntohl(myip.s_addr)));

	if (target->ip == htonl(LUNAR_IP_BCAST)) {
		unsigned char buf[300], *cp = buf;
		sapfsel_t ptr;

		DEBUG(1, printk("broadcast tree request received\n"));

		if (!fwd) {
			DEBUG(9, printk(" no fwd sel\n"));
			return;
		}
		ptr = SAPF_Instance sapf_getUnusedSelector(mysapfhandle, fwd->sel);
		SAPF_Instance sapf_addREDIR(mysapfhandle, ptr, ipv4deliver);

		cp += xrp_fill_hdr(cp, XRP_VERSION_REPLY, 0);
		cp += xrp_fill_sel_eth(cp, XRP_CLASS_FWD_ADDR, ptr, (unsigned char *)sa.sa_data);
		unsigned char hid[NETBOX_HID_LEN];

		if (!NETBOX_Instance netbox_getHID128(hid))
			cp += xrp_fill_hid128(cp, XRP_CLASS_SRC_ID, hid);
		*cp++ = 0; *cp++ = 0;

		memcpy(sa.sa_data, reply->eth, ETH_ALEN);
		NETBOX_Instance netbox_l2send(ifnr, &sa, reply->sel, buf, cp - buf);

		return;
	}

	if (target->ip == myip.s_addr) { // unicast match
		unsigned char buf[300], *cp = buf;
		unsigned char hid[NETBOX_HID_LEN];
		sapfsel_t ptr = SAPF_Instance sapf_getUnusedSelector(mysapfhandle,
						       SELCLASS_LOCAL);

		DEBUG(10, printk("LUNAR: I was discovered\n"));

		// fprintf(stderr, "backwd=%p src=%p, %08x\n", backwd, src, src ? src->ip : 0);
		if (backwd && src && peer_is_unknown(src->ip)) {
			struct in_addr ia;
			ia.s_addr = src->ip;
			install_peer(ia, mkNetPointer(backwd->eth, backwd->sel), 9, 1);
		}

		SAPF_Instance sapf_addREDIR(mysapfhandle, ptr, ipv4deliver);

		cp += xrp_fill_hdr(cp, XRP_VERSION_REPLY, 0);
		cp += xrp_fill_sel_eth(cp, XRP_CLASS_FWD_ADDR, ptr, (unsigned char*)sa.sa_data);

		if (!NETBOX_Instance netbox_getHID128(hid))
			cp += xrp_fill_hid128(cp, XRP_CLASS_SRC_ID, hid);
		*cp++ = 0; *cp++ = 0;

		memcpy(sa.sa_data, reply->eth, ETH_ALEN);
		NETBOX_Instance netbox_l2send(ifnr, &sa, reply->sel, buf, cp - buf);
	} else if (x->ttl-- > 1) { // re-broadcast the request
		struct uc_transit_pending_s *tr;

		if (!(tr = (struct uc_transit_pending_s *)kmalloc(sizeof(*tr), GFP_ATOMIC)))
			return;
		memset(tr, 0, sizeof(*tr));

		tr->replysel = reply->sel;
		memcpy(tr->replyeth, reply->eth, ETH_ALEN);

		memcpy(reply->eth, sa.sa_data, ETH_ALEN);
		tr->caller = reply->sel = mkHandlerPointer(
			NS_CLASS_addr unicast_transit_rrep_handler, tr);
		if (backwd) {
			tr->is_bidir = 1;
			tr->backsel = backwd->sel;
			memcpy(tr->backeth, backwd->eth, ETH_ALEN);

			tr->mybacksel = backwd->sel =
				SAPF_Instance sapf_getUnusedSelector(mysapfhandle, SELCLASS_LOCAL);
			memcpy(backwd->eth, sa.sa_data, ETH_ALEN);
		}

		schedule_rreq(p->data, p->len, 0);
	} // else drop

	return;
}


void NS_CLASS send_rreq(int radius, sapfsel_t series, uint32_t target,
	       sapfsel_t reply, sapfsel_t backsel, int32_t src,
	       sapfsel_t fwd)
{
	struct sockaddr sa;
	int ifnr = basedev ? basedev->ifindex : 0;
	char buf[300], *cp = buf;

	cp = buf;
	cp += xrp_fill_hdr(cp, XRP_VERSION_REQUEST, radius);
	cp += xrp_fill_sel(cp, XRP_CLASS_REQ_SERIES, series);
	NETBOX_Instance netbox_getL2addr(ifnr, &sa);
	cp += xrp_fill_sel_eth(cp, XRP_CLASS_REPLY_ADDR,
			       reply, (unsigned char *)sa.sa_data);
	cp += xrp_fill_ipv4(cp, XRP_CLASS_TARGET, target);

	if (src != NO_IP_ADDR) {
		cp += xrp_fill_ipv4(cp, XRP_CLASS_SRC_ADDR, src);
		cp += xrp_fill_sel_eth(cp, XRP_CLASS_BACKWD_ADDR,
			       backsel, (unsigned char *)sa.sa_data);
	}
	// else fprintf(stderr, "send_rreq: no src\n");

	if (fwd) {
		cp += xrp_fill_sel(cp, XRP_CLASS_FWD_ADDR, fwd);
	}

	*cp++ = 0; *cp++ = 0;

	NETBOX_Instance netbox_l2send(ifnr, 0,
				      htonsel(SAPF_SELNET_LUNAR_RREQ),
				      buf, cp - buf);
}

 void NS_CLASS start_rreq(int ndx)
{
	sapfsel_t series;
	int radius, i;
	long tout;

	DEBUG(10, printk("start_rreq called for %08x\n", peers[ndx].ia.s_addr));

	radius = MIN(peers[ndx].used, conf_max_ring);
	series = SAPF_Instance sapf_getUnusedSelector(mysapfhandle, SELCLASS_REMOTE);
	SAPF_Instance sapf_addREDIR(mysapfhandle, series, 0);

	if (peers[ndx].ia.s_addr != htonl(LUNAR_IP_BCAST)) {
		struct uc_pending_s *ups;

		ups = (struct uc_pending_s *)kmalloc(sizeof(struct uc_pending_s), GFP_ATOMIC);
		if (!ups) {
			peers[ndx].used = 0;
			return;
		}
		memset(ups, 0, sizeof(struct uc_pending_s));
		ups->ia = peers[ndx].ia;
		ups->caller = mkHandlerPointer(
			NS_CLASS_addr unicast_rrep_handler, ups);

		if (peers[ndx].flags & PEER_BIDIR) {
			struct in_addr myip;

			ups->is_bidir = 1;
			ups->backsel = peers[ndx].backsel;
			NETBOX_Instance netbox_getL3addr(&myip, lunardev);
			send_rreq(radius, series, peers[ndx].ia.s_addr,
				  ups->caller, ups->backsel, myip.s_addr, 0);
		} else
			send_rreq(radius, series, peers[ndx].ia.s_addr,
				  ups->caller, 0, NO_IP_ADDR, 0);

		tout = radius * conf_timeout_per_ring;
		i = peers[ndx].used;
		/* cft: removed after having written the ID
		if (i > conf_max_ring) {
		  	tout += conf_retry_wait;
			while (--i > conf_max_ring)
				tout *= 2;
		}
		*/
	} else { // set up broadcast tree
		struct bcast_pending_s *bps;

		bps = (struct bcast_pending_s *)kmalloc(sizeof(struct bcast_pending_s), GFP_ATOMIC);
		if (!bps) {
			peers[ndx].used = 0;
			return;
		}
		memset(bps, 0, sizeof(struct bcast_pending_s));
		bps->caller = mkHandlerPointer(
			NS_CLASS_addr bcast_rep_handler, bps);
		bps->fwd_sel = SAPF_Instance sapf_getUnusedSelector(mysapfhandle, SELCLASS_REMOTE);
		bps->ia = peers[ndx].ia;
		send_rreq(radius, series, bps->ia.s_addr,
			  bps->caller, 0, NO_IP_ADDR, bps->fwd_sel);
		if (bc_len)
			tout = LUNAR_BRCAST_DELAY;
		else
			tout = conf_fri;
	}

	NETBOX_Instance netbox_newtimer(&(peers[ndx].timer), _VOID_VOID_ peer_timeout,
			peers + ndx, tout);
}


 void NS_CLASS peer_timeout(void *vp)
{
  struct peer_s *n=(struct peer_s *)vp;
	DEBUG(9, printk("peer_timeout %08x (used=%d)\n",
			ntohl(n->ia.s_addr), n->used));

	if (n->used == 0)
		return;

	if (n->used == -1) {
		if (SAPF_Instance sapf_testandreset_active(mysapfhandle, n->fwdsel)) {
			DEBUG(10, printk("restarting rreq because of activity\n"));
			if (n->ia.s_addr == htonl(LUNAR_IP_BCAST))
				n->used = (short) conf_max_retries;
			else
				n->used = (short) conf_min_ring;
			start_rreq(n - peers);
		} else {
			DEBUG(10, printk("nonactive peer timeout\n"));
			n->used = 0;
			SAPF_Instance sapf_deleteEntry(mysapfhandle, n->fwdsel);
			if (n->ia.s_addr == htonl(LUNAR_IP_BCAST)) {
				DEBUG(10, printk("  bc_selp %p\n", bc_selp));
				bc_sel = 0;
				if (bc_selp) {
					*bc_selp = 0;
					bc_selp = 0;
				}
			} else
				NETBOX_Instance netbox_arpdelete(n->ia, lunardev);
		}
		return;
	}

	if (n->ia.s_addr == htonl(LUNAR_IP_BCAST)) {
		DEBUG(9, printk("bc_delayed_start\n"));

		if (bc_len) {
			if (bc_sel) {
				sapfpkt pkt;

				pkt.data = (char*) bc_buf;
				pkt.len = bc_len;
				SAPF_Instance sapf_demux(mysapfhandle, bc_sel, &pkt);
			}
			bc_len = 0;
		}
		if (bc_selp)
			*bc_selp = bc_sel;
		if (bc_sel)
			install_peer(n->ia, bc_sel, 3, 0);
		
		return;
	}

	if (n->used >= conf_max_retries) { // we give up
		DEBUG(9, printk("LUNAR path setup failed\n"));
		n->used = 0;
		NETBOX_Instance netbox_arpdelete(n->ia, lunardev);
		return;
	}

	if (n->used < conf_max_ring)
		// 1-hop failed, so try immediately max hops
		n->used = (short) conf_max_ring;
	else    // count up to conf_max_retries
		n->used ++;
	start_rreq(n - peers);
}


// ----------------------------------------------------------------------

 int NS_CLASS lunar_do_arp(struct in_addr name)
{
	int ndx;

	DEBUG(10, printk("lunar ARP  %08x\n", ntohl(name.s_addr)));

	ndx = add_peer(name);
	if (ndx == -1)
		return 0;
	if (peers[ndx].used == -1) {
		arp_answer(ndx);
		return 0;
	}
//	peers[ndx].flags = PEER_ARP_PENDING;
//	netbox_arpannounce(name, lunardev);

	if (peers[ndx].used > 0) // resolution is pending
		return -1;
	peers[ndx].used = (short) conf_min_ring;
	peers[ndx].backsel = SAPF_Instance sapf_getUnusedSelector(mysapfhandle, SELCLASS_REMOTE);
	peers[ndx].flags |= PEER_BIDIR;

	start_rreq(ndx);

	return -1;
}


 void NS_CLASS lunar_do_bc_setup(sapfsel_t *sel, sapfpkt *pkt)
{
	struct in_addr bc;
	int ndx;

	DEBUG(9, printk("lunar_do_bc_setup called\n"));

	bc.s_addr = htonl(LUNAR_IP_BCAST);
	ndx = add_peer(bc);
	if (ndx == -1 || peers[ndx].used != 0) {
		return;
	}

	memcpy(bc_buf, pkt->data, pkt->len);
	bc_len = pkt->len;
	bc_sel = 0;
	bc_selp = sel;
	
	peers[ndx].used = (short) conf_max_retries;
	peers[ndx].backsel = SAPF_Instance sapf_getUnusedSelector(mysapfhandle, SELCLASS_REMOTE);
	peers[ndx].flags |= PEER_BCTREE;

	start_rreq(ndx);
}

// ----------------------------------------------------------------------

int NS_CLASS lunar_proc_info(char *buf, char **start, off_t offset, int len)
{
	int written = 0, i, cnt;
	struct peer_s *p;

	for (p = peers, i = LUNAR_MAX_PEERS, cnt = 0; i > 0; p++, i--)
		if (p->used)
			cnt++;

	written += sprintf(buf, "# lunar (%d table entries, %d %s)\n", LUNAR_MAX_PEERS, cnt, cnt == 1 ? "peer" : "peers");

#define adjustlen()	{ \
	if (written < offset) { offset -= written; written = 0; } \
	if (len <= written) \
		goto out; \
}
	for (p = peers, i = 0; i < LUNAR_MAX_PEERS; i++, p++) {
		char tmp[20];
		uint32_t ip = ntohl(p->ia.s_addr);
		if (!p->used)
			continue;
		// in kernel space we don't have inet_ntoa():
		sprintf(tmp, "%d.%d.%d.%d",
		    0x0ff & (ip >> 24), 0x0ff & (ip >> 16),
		    0x0ff & (ip >> 8),  0x0ff & ip);
		written += sprintf(buf + written, "%-15s  ", tmp);
		if (p->used > 0) {
			written += sprintf(buf + written,
					   "   searching (attempt %d)\n", p->used);
		} else {
			struct sapfdest_s *d;
			unsigned char *cp;

			written += sprintf(buf + written, p->hops == 0x0f ? 
					   " ? " : "%2d ", p->hops+1);

			d = (struct sapfdest_s*)
				SAPF_Instance sapf_getCallbackAux(mysapfhandle, p->fwdsel);
			cp = (unsigned char *)d->dst.sa_data;
			written += sprintf(buf + written,
					   "%02x:%02x:%02x:%02x:%02x:%02x / %016Lx\n",
					   cp[0], cp[1], cp[2], cp[3],
					   cp[4], cp[5], ntohsel(d->sel));
		}
		adjustlen();
	}

out:
	*start = buf + offset;
	written -= offset;
	if (written > len)
		return len;
	return (written < 0) ? 0 : written ;
}


// eof
