/* lunar_init.c
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   generic LUNAR init functionality
 *
 * Copyright (C) 2001-2004 <christian.tschudin@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 *
 *   2004-01-09 extracted from linux/klunar.c
 */

 void NS_CLASS L2sendto(sapfpkt *pkt, void *aux)
{
	struct sapfdest_s *dst = (struct sapfdest_s *) aux;

	if (!aux)
		return;

	DEBUG(10, printk("L2send %d bytes\n", pkt->len));

	NETBOX_Instance netbox_l2send(dst->ifnr, &(dst->dst),
				      dst->sel, pkt->data, pkt->len);
	return;
}

void NS_CLASS L3deliver(sapfpkt *pkt, void *aux)
{
#ifndef NS_PORT
        struct net_device_stats *stats;

	stats = (struct net_device_stats *) lunardev->priv;
#endif	
	DEBUG(10, printk("L3deliver %d Bytes\n", pkt->len));

	if (stats) {
		stats->rx_packets++;
		stats->rx_bytes += pkt->len;
	}

	//	netbox_l3deliver(pkt->ifnr, pkt->data, pkt->len);
	NETBOX_Instance netbox_l3deliver(lunardev->ifindex, pkt->data, pkt->len);

	return;
}


// ----------------------------------------------------------------------

int NS_CLASS lunar_init(void)
{
	DEBUG(1, printk("loading lunar driver\n"));

	dgwmode = forwarding && !strcmp(forwarding, "yes");

	mysapfhandle = SAPF_Instance sapf_open("main", 641);
	if (mysapfhandle < 0) {
		DEBUG(10, printk("sapf_open failed\n"));
		return -1;
	}

	if (NETBOX_Instance netbox_init()) {
		SAPF_Instance sapf_close(mysapfhandle);
		return -1;
	}

	SAPF_Instance sapf_addCallback(mysapfhandle, htonsel(SAPF_SELNET_LUNAR_RREQ),
				       NS_CLASS_addr lunar_rreq_handler, 0);
	SAPF_Instance sapf_setPermanent(mysapfhandle, htonsel(SAPF_SELNET_LUNAR_RREQ));

	ipv4deliver = SAPF_Instance sapf_getUnusedSelector(mysapfhandle, SELCLASS_LOCAL);
	SAPF_Instance sapf_addCallback(mysapfhandle, ipv4deliver, NS_CLASS_addr L3deliver, 0);
	SAPF_Instance sapf_setPermanent(mysapfhandle, ipv4deliver);

	return 0;
}


void NS_CLASS lunar_exit(void)
{
	DEBUG(1, printk("unloading lunar driver\n"));

	SAPF_Instance sapf_deleteEntry(mysapfhandle, htonsel(SAPF_SELNET_LUNAR_RREQ));
	SAPF_Instance sapf_deleteEntry(mysapfhandle, ipv4deliver);
	ipv4deliver = 0;

	if (basedev) {
		struct net_device *d = basedev;
		basedev = 0;
		NETBOX_Instance netbox_dropL2if(d->ifindex);
	}

	NETBOX_Instance netbox_fini();

	SAPF_Instance sapf_close(mysapfhandle);
}

// eof
