/* passthru.c */

#include "precomp.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "sapf.h"
#include "lunar.h"

/* Size of static data holder (in bytes) */
#define STATIC_DATA_SIZE 1000

extern int sapf_init(void);
extern void sapf_fini(void);
extern int klunar_init(void);
extern void klunar_fini(void);

extern void activate_packet_pump(PADAPT pAdapt);

/* Defined in netbox module */
extern NDIS_SPIN_LOCK l2rcvLock;
extern LIST_ENTRY l2rcvQueue;
extern NDIS_SPIN_LOCK l2intrvOutLock;
extern LIST_ENTRY l2intrvOutQueue;

BOOLEAN lunar_init_done = FALSE;
/* Can be changed to FALSE to boot before activating */
BOOLEAN TimeToInitLunar = FALSE;
PADAPT StoredAdapter = NULL;

/* Static data holder used to save time */
static BYTE MyStaticData[STATIC_DATA_SIZE];

extern BYTE* GetMACAddress();
extern void send_notification(unsigned long event);

#pragma hdrstop

#pragma NDIS_INIT_FUNCTION(DriverEntry)

NDIS_HANDLE ProtHandle = NULL;
NDIS_HANDLE DriverHandle = NULL;
NDIS_MEDIUM MediumArray[4] =
{
  NdisMedium802_3, // Ethernet
  NdisMedium802_5, // Token-ring
  NdisMediumFddi,  // Fddi
  NdisMediumWan    // NDISWAN
};

NDIS_SPIN_LOCK GlobalLock;

PADAPT	       pAdaptList = NULL;
LONG           MiniportCount = 0;

NDIS_HANDLE    NdisWrapperHandle;

//
// To support ioctls from user-mode:
//

#define LINKNAME_STRING     L"\\DosDevices\\Passthru"
#define NTDEVICE_STRING     L"\\Device\\Passthru"

NDIS_HANDLE     NdisDeviceHandle = NULL;
PDEVICE_OBJECT  ControlDeviceObject = NULL;

enum _DEVICE_STATE
{
  PS_DEVICE_STATE_READY = 0,	// ready for create/delete
  PS_DEVICE_STATE_CREATING,	// create operation in progress
  PS_DEVICE_STATE_DELETING	// delete operation in progress
} ControlDeviceState = PS_DEVICE_STATE_READY;

PADAPT GetAdapter() {
  return StoredAdapter;
}

VOID InitLunar() {
  TimeToInitLunar = TRUE;
}

VOID HaltLunar() {
  if (lunar_init_done) {
    lunar_init_done = FALSE;
    TimeToInitLunar = FALSE;

    /* Note: Done here to be symmetric with the NETDEV_UP */
    send_notification(NETDEV_DOWN);
    // TODO: Make the finalization work as well
    // without crashing Windows
    klunar_fini(); // This in turn calls netbox_fini()
    sapf_fini();
    platform_fini();
  }
}

VOID DumpPacket(IN BYTE* data, IN UINT length, IN UINT PacketType)
{
  BYTE* MyString;
  BYTE* MyString2;
  BYTE* MyBuffer;
  BYTE* MyBuffer2;
  UINT i;
  UINT OffsetPayload;

  MyString = (BYTE*)kmalloc(length * 4, 0);
  MyString2 = (BYTE*)kmalloc(length * 4, 0);
  MyBuffer = (BYTE*)kmalloc(128, 0);
  MyBuffer2 = (BYTE*)kmalloc(128, 0);

  if (!MyString || !MyString2 || !MyBuffer || !MyBuffer2) {
    if (MyString) kfree(MyString);
    if (MyString2) kfree(MyString2);
    if (MyBuffer) kfree(MyBuffer);
    if (MyBuffer2) kfree(MyBuffer2);
    return;
  }

  // Print some info about the packet
  if (PacketType == PT_IP) {
    DbgPrint("IP Packet, length: %d\n", length);
  } else if (PacketType == PT_ARP) {
    DbgPrint("ARP Packet, length: %d\n", length);
  } else if (PacketType == PT_SAPF) {
    DbgPrint("SAPF Packet, length: %d\n", length);
  }

  // Print MAC addresses of sender and receiver
  MyString[0] = '\0';
  for (i = 6; i < 12; i++) {
    sprintf(MyBuffer, "%02X ", data[i]);
    strcat(MyString, MyBuffer);
  }
  DbgPrint("Source MAC: %s", MyString);
  MyString[0] = '\0';
  for (i = 0; i < 6; i++) {
    sprintf(MyBuffer, "%02X ", data[i]);
    strcat(MyString, MyBuffer);
  }
  DbgPrint("Destination MAC: %s", MyString);

  // If IP packet, print IP addresses and
  // calculate offset to payload
  if (PacketType == PT_IP) {
    MyString[0] = '\0';
    for (i = 26; i < 30; i++) {
      sprintf(MyBuffer, "%d.", data[i]);
      strcat(MyString, MyBuffer);
    }
    DbgPrint("Source IP: %s", MyString);
    MyString[0] = '\0';
    for (i = 30; i < 34; i++) {
      sprintf(MyBuffer, "%d.", data[i]);
      strcat(MyString, MyBuffer);
    }
    DbgPrint("Destination IP: %s", MyString);

    //OffsetPayload = ETH_HLEN + (data[ETH_HLEN] & 0x0F);
    OffsetPayload = 0;
    // If ARP or SAPF packet, just calculate offset to payload
  } else if ((PacketType == PT_ARP) || (PacketType == PT_SAPF)) {
    OffsetPayload = ETH_HLEN;
  }

  /* Truncate printed data in order not to
     DbgPrint too much */
  if (length > (OffsetPayload + 128)) {
    length = (OffsetPayload + 128);
  }

  MyString[0] = '\0';
  MyString2[0] = '\0';
  for (i = OffsetPayload; i < length; i++) {
    sprintf(MyBuffer, "%02X ", data[i]);
    strcat(MyString, MyBuffer);

    sprintf(MyBuffer2, "%c", data[i]);
    strcat(MyString2, MyBuffer2);
  }

  DbgPrint("Packet contents: %s", MyString);
  DbgPrint("Packet chars: %s", MyString2);

  kfree(MyString);
  kfree(MyString2);
  kfree(MyBuffer);
  kfree(MyBuffer2);
}


NTSTATUS
DriverEntry(
	    IN PDRIVER_OBJECT  DriverObject,
	    IN PUNICODE_STRING RegistryPath
	    )
     /*++

Routine Description:

	First entry point to be called, when this driver is loaded.
	Register with NDIS as an intermediate driver.

Arguments:

	DriverObject - pointer to the system's driver object structure
		       for this driver
	
	RegistryPath - system's registry path for this driver
	
Return Value:

	STATUS_SUCCESS if all initialization is successful, STATUS_XXX
	error code if not.

--*/
{
  NDIS_STATUS			Status;
  NDIS_PROTOCOL_CHARACTERISTICS	PChars;
  NDIS_MINIPORT_CHARACTERISTICS	MChars;
  PNDIS_CONFIGURATION_PARAMETER	Param;
  NDIS_STRING			Name;

  Status = NDIS_STATUS_SUCCESS;
  NdisAllocateSpinLock(&GlobalLock);

  NdisMInitializeWrapper(&NdisWrapperHandle, DriverObject, RegistryPath, NULL);

  do
    {
      //
      // Register the miniport with NDIS. Note that it is the miniport
      // which was started as a driver and not the protocol. Also the miniport
      // must be registered prior to the protocol since the protocol's BindAdapter
      // handler can be initiated anytime and when it is, it must be ready to
      // start driver instances.
      //

      NdisZeroMemory(&MChars, sizeof(NDIS_MINIPORT_CHARACTERISTICS));

      MChars.MajorNdisVersion = PASSTHRU_MAJOR_NDIS_VERSION;
      MChars.MinorNdisVersion = PASSTHRU_MINOR_NDIS_VERSION;

      MChars.InitializeHandler = MPInitialize;
      MChars.QueryInformationHandler = MPQueryInformation;
      MChars.SetInformationHandler = MPSetInformation;
      MChars.ResetHandler = MPReset;
      MChars.TransferDataHandler = MPTransferData;
      MChars.HaltHandler = MPHalt;
#ifdef NDIS51_MINIPORT
      MChars.CancelSendPacketsHandler = MPCancelSendPackets;
      MChars.PnPEventNotifyHandler = MPDevicePnPEvent;
      MChars.AdapterShutdownHandler = MPAdapterShutdown;
#endif // NDIS51_MINIPORT

      //
      // We will disable the check for hang timeout so we do not
      // need a check for hang handler!
      //
      MChars.CheckForHangHandler = NULL;
      MChars.ReturnPacketHandler = MPReturnPacket;

      //
      // Either the Send or the SendPackets handler should be specified.
      // If SendPackets handler is specified, SendHandler is ignored
      //
      MChars.SendHandler = NULL;	// MPSend;
      MChars.SendPacketsHandler = MPSendPackets; // NULL;

      Status = NdisIMRegisterLayeredMiniport(NdisWrapperHandle,
					     &MChars,
					     sizeof(MChars),
					     &DriverHandle);
      if (Status != NDIS_STATUS_SUCCESS)
	{
	  break;
	}

#ifndef WIN9X
      NdisMRegisterUnloadHandler(NdisWrapperHandle, PtUnload);
#endif

      //
      // Now register the protocol.
      //
      NdisZeroMemory(&PChars, sizeof(NDIS_PROTOCOL_CHARACTERISTICS));
      PChars.MajorNdisVersion = PASSTHRU_PROT_MAJOR_NDIS_VERSION;
      PChars.MinorNdisVersion = PASSTHRU_PROT_MINOR_NDIS_VERSION;

      //
      // Make sure the protocol-name matches the service-name
      // (from the INF) under which this protocol is installed.
      // This is needed to ensure that NDIS can correctly determine
      // the binding and call us to bind to miniports below.
      //
      NdisInitUnicodeString(&Name, L"Passthru");	// Protocol name
      PChars.Name = Name;
      PChars.OpenAdapterCompleteHandler = PtOpenAdapterComplete;
      PChars.CloseAdapterCompleteHandler = PtCloseAdapterComplete;
      PChars.SendCompleteHandler = PtSendComplete;
      PChars.TransferDataCompleteHandler = PtTransferDataComplete;
	
      PChars.ResetCompleteHandler = PtResetComplete;
      PChars.RequestCompleteHandler = PtRequestComplete;
      PChars.ReceiveHandler = PtReceive;
      PChars.ReceiveCompleteHandler = PtReceiveComplete;
      PChars.StatusHandler = PtStatus;
      PChars.StatusCompleteHandler = PtStatusComplete;
      PChars.BindAdapterHandler = PtBindAdapter;
      PChars.UnbindAdapterHandler = PtUnbindAdapter;
      PChars.UnloadHandler = PtUnloadProtocol;

      PChars.ReceivePacketHandler = PtReceivePacket;
      PChars.PnPEventHandler= PtPNPHandler;

      NdisRegisterProtocol(&Status,
			   &ProtHandle,
			   &PChars,
			   sizeof(NDIS_PROTOCOL_CHARACTERISTICS));

      if (Status != NDIS_STATUS_SUCCESS)
	{
	  NdisIMDeregisterLayeredMiniport(DriverHandle);
	  break;
	}

      NdisIMAssociateMiniport(DriverHandle, ProtHandle);
    }
  while (FALSE);

  if (Status != NDIS_STATUS_SUCCESS)
    {
      NdisTerminateWrapper(NdisWrapperHandle, NULL);
    }

  return(Status);
}


NDIS_STATUS
PtRegisterDevice(
		 VOID
		 )
     /*++

Routine Description:

	Register an ioctl interface - a device object to be used for this
	purpose is created by NDIS when we call NdisMRegisterDevice.

	This routine is called whenever a new miniport instance is
	initialized. However, we only create one global device object,
	when the first miniport instance is initialized. This routine
	handles potential race conditions with PtDeregisterDevice via
	the ControlDeviceState and MiniportCount variables.

	NOTE: do not call this from DriverEntry; it will prevent the driver
	from being unloaded (e.g. on uninstall).

Arguments:

	None

Return Value:

	NDIS_STATUS_SUCCESS if we successfully register a device object.

--*/
{
  NDIS_STATUS	   Status = NDIS_STATUS_SUCCESS;
  UNICODE_STRING   DeviceName;
  UNICODE_STRING   DeviceLinkUnicodeString;
  PDRIVER_DISPATCH DispatchTable[IRP_MJ_MAXIMUM_FUNCTION];
  UINT		   i;

  DBGPRINT(("==>PtRegisterDevice\n"));

  NdisAcquireSpinLock(&GlobalLock);

  ++MiniportCount;
	
  if (1 == MiniportCount)
    {
      ASSERT(ControlDeviceState != PS_DEVICE_STATE_CREATING);

      //
      // Another thread could be running PtDeregisterDevice on
      // behalf of another miniport instance. If so, wait for
      // it to exit.
      //
      while (ControlDeviceState != PS_DEVICE_STATE_READY)
	{
	  NdisReleaseSpinLock(&GlobalLock);
	  NdisMSleep(1);
	  NdisAcquireSpinLock(&GlobalLock);
	}

      ControlDeviceState = PS_DEVICE_STATE_CREATING;

      NdisReleaseSpinLock(&GlobalLock);

      for (i = 0; i < IRP_MJ_MAXIMUM_FUNCTION; i++)
	{
	  DispatchTable[i] = PtDispatch;
	}

      NdisInitUnicodeString(&DeviceName, NTDEVICE_STRING);
      NdisInitUnicodeString(&DeviceLinkUnicodeString, LINKNAME_STRING);

      //
      // Create a device object and register our dispatch handlers
      //
		
      Status = NdisMRegisterDevice(
				   NdisWrapperHandle, 
				   &DeviceName,
				   &DeviceLinkUnicodeString,
				   &DispatchTable[0],
				   &ControlDeviceObject,
				   &NdisDeviceHandle
				   );

      NdisAcquireSpinLock(&GlobalLock);

      ControlDeviceState = PS_DEVICE_STATE_READY;
    }

  NdisReleaseSpinLock(&GlobalLock);

  DBGPRINT(("<==PtRegisterDevice: %x\n", Status));

  /* Send notification about the event to Lunar */
  /* TODO: Perhaps we should do this; Then also rework
     the device handling and make it more general */
  /* send_notification(NETDEV_REGISTER); */

  return (Status);
}


NTSTATUS
PtDispatch(
	   IN PDEVICE_OBJECT DeviceObject,
	   IN PIRP	     Irp
	   )
     /*++
Routine Description:

	Process IRPs sent to this device.

Arguments:

	DeviceObject - pointer to a device object
	Irp	     - pointer to an I/O Request Packet

Return Value:

	NTSTATUS - STATUS_SUCCESS always - change this when adding
	real code to handle ioctls.

--*/
{
  PIO_STACK_LOCATION irpStack;
  NTSTATUS	     status = STATUS_SUCCESS;

  DBGPRINT(("==>Pt Dispatch\n"));
  irpStack = IoGetCurrentIrpStackLocation(Irp);

  switch (irpStack->MajorFunction)
    {
    case IRP_MJ_CREATE:
      break;
    case IRP_MJ_CLOSE:
      break;        
    case IRP_MJ_DEVICE_CONTROL:
      //
      // Add code here to handle ioctl commands sent to passthru.
      //
      break;        
    default:
      break;
    }

  Irp->IoStatus.Status = status;
  IoCompleteRequest(Irp, IO_NO_INCREMENT);

  DBGPRINT(("<== Pt Dispatch\n"));

  return status;
} 


NDIS_STATUS
PtDeregisterDevice(
		   VOID
		   )
     /*++

Routine Description:

	Deregister the ioctl interface. This is called whenever a miniport
	instance is halted. When the last miniport instance is halted, we
	request NDIS to delete the device object

Arguments:

	NdisDeviceHandle - Handle returned by NdisMRegisterDevice

Return Value:

	NDIS_STATUS_SUCCESS if everything worked ok

--*/
{
  NDIS_STATUS Status = NDIS_STATUS_SUCCESS;

  DBGPRINT(("==>PassthruDeregisterDevice\n"));

  NdisAcquireSpinLock(&GlobalLock);

  ASSERT(MiniportCount > 0);

  --MiniportCount;
	
  if (0 == MiniportCount)
    {
      //
      // All miniport instances have been halted. Deregister
      // the control device.
      //

      ASSERT(ControlDeviceState == PS_DEVICE_STATE_READY);

      //
      // Block PtRegisterDevice() while we release the control
      // device lock and deregister the device.
      // 
      ControlDeviceState = PS_DEVICE_STATE_DELETING;

      NdisReleaseSpinLock(&GlobalLock);

      if (NdisDeviceHandle != NULL)
	{
	  Status = NdisMDeregisterDevice(NdisDeviceHandle);
	  NdisDeviceHandle = NULL;
	}

      NdisAcquireSpinLock(&GlobalLock);
      ControlDeviceState = PS_DEVICE_STATE_READY;
    }

  NdisReleaseSpinLock(&GlobalLock);

  DBGPRINT(("<== PassthruDeregisterDevice: %x\n", Status));

  /* Send notification about the event to Lunar */
  /* TODO: Perhaps we should do this; Then also rework
     the device handling and make it more general */
  /* send_notification(NETDEV_UNREGISTER); */

  return Status;	
}


VOID
PtUnloadProtocol(
		 VOID
		 )
{
  NDIS_STATUS Status;

  if (ProtHandle != NULL)
    {
      NdisDeregisterProtocol(&Status, ProtHandle);
      ProtHandle = NULL;
    }

  DBGPRINT(("PtUnloadProtocol: done!\n"));
}


/* Returns TRUE if the given packet was "consumed"
   by this function and does not need to be handled
   further */
BOOLEAN
AnalyzePacketBuffer(
		    IN BYTE* PacketBuffer,
		    IN UINT TotalPacketLength,
		    IN COMDIRECTION Direction,
		    IN BOOLEAN VolatileData,
		    IN PADAPT pAdapt)
{
  UINT PacketType = PT_UNKNOWN;
  struct sk_buff *skb = NULL;
  PQUEUED_PACKET packetToLunar = NULL;
  BYTE* dataCopy;

  // Disregard from packets that cannot be IP, ARP or
  // SAPF; Just an initial distinguishing operation
  // (to increase performance)
  // TODO: Perhaps restore later
  //if (TotalPacketLength < 32) {
  /* Indicate for the packet to be handled
     elsewhere */
  //return FALSE;
  //}

  /* Identify packet type */
  PacketType = PT_UNKNOWN;
  switch (ntohs(*(unsigned short*)(PacketBuffer + 12)))
    {
    case ETH_P_IP:
      PacketType = PT_IP;
      break;
    case ETH_P_ARP:
      PacketType = PT_ARP;
      break;
    case SAPF_ETHERTYPE:
      PacketType = PT_SAPF;
      break;
    }

  // We don't handle anything other than IP, ARP
  // and SAPF packets
  // TODO: Perhaps restore later
  //if (PacketType == PT_UNKNOWN) {
  //DbgPrint("Packet type unknown/unhandled\n");
  /* Indicate for the packet to be handled
     elsewhere */
  //return TRUE;
  //}

  /* Note: Here, blinding of traffic from a certain
     MAC address can be done */
  /*
  if (PacketBuffer[6] == 0x00 &&
      PacketBuffer[7] == 0x0c &&
      PacketBuffer[8] == 0x29 &&
      PacketBuffer[9] == 0x29 &&
      PacketBuffer[10] == 0x43 &&
      PacketBuffer[11] == 0xBF)
    {
      DbgPrint("Blinding traffic from node 12");
      return TRUE;
    }
  */

  /* We init Lunar when the first packet is to
     be handled */
  if (TimeToInitLunar && !lunar_init_done && pAdapt != NULL) {
    platform_init();
    /* In order to use this adapter for the other
       initalization */
    StoredAdapter = pAdapt;
    sapf_init();
    klunar_init(); /* This in turn calls netbox_init() */
    /* Note: Done here to e.g. init handle */
    send_notification(NETDEV_UP);
    lunar_init_done = TRUE;
    DbgPrint("Done performing WinLunar initialization");
  }

  /* If the initialization of Lunar has not
     yet been performed, return and let Windows
     worry about the packet itself */
  /* TODO: Can be removed later (to increase performance) */
  if (!lunar_init_done) {
    return FALSE;
  }

  /* If the packet analyzed is an outgoing IP or
     ARP packet, we intervene it and send to Lunar
     for handling */
  if ((Direction == Send) && ((PacketType == PT_ARP) ||
			      (PacketType == PT_IP))) {

    /* If the given data is "volatile", then we need
       to make a copy of the data before sending it on */
    if (!VolatileData) {
      dataCopy = PacketBuffer;
    } else {
      dataCopy = kmalloc(TotalPacketLength, 0);
      if (dataCopy) {
	memcpy(dataCopy, PacketBuffer, TotalPacketLength);
      }
    }

    if (dataCopy) {
      packetToLunar = kmalloc(sizeof(QUEUED_PACKET), 0);
      if (packetToLunar) {
	packetToLunar->buffer = dataCopy;
	packetToLunar->length = TotalPacketLength;
	packetToLunar->direction = NDIS_DIR_SEND; // Not used
	NdisInitializeListHead(&packetToLunar->listEntry);
	/* Note: Allocated packet memory is freed when
	   the list entry is removed and handled */
	NdisInterlockedInsertTailList(&l2intrvOutQueue,
				      &packetToLunar->listEntry, &l2intrvOutLock);
	
	/* Indicate that a handler thread should take
	   the packet */
	activate_packet_pump(pAdapt);
      }
    }

    /* Indicate that the packet can be dropped (in order
       to reduce outgoing network traffic) */
    return TRUE;
  }

  /* If the packet analyzed is an incoming SAPF packet, then
     we construct the necessary data objects and call the
     l2recv function in the Lunar netbox module */
  if ((Direction == Receive) && (PacketType == PT_SAPF)) {
    skb = kmalloc(sizeof(struct sk_buff), 0);
    if (skb) {
      /* If the given data is "volatile", then we need
	 to make a copy of the data before sending it on */
      if (!VolatileData) {
	dataCopy = PacketBuffer + ETH_HLEN;
      } else {
	dataCopy = kmalloc(TotalPacketLength - ETH_HLEN, 0);
	if (dataCopy) {
	  memcpy(dataCopy, PacketBuffer + ETH_HLEN, TotalPacketLength - ETH_HLEN);
	}
      }

      if (dataCopy) {
	skb->nh.raw = dataCopy;
	/* The control buffer we just leave as is */
	/* skb->cb[0] = ... */
	skb->len = TotalPacketLength - ETH_HLEN;
	skb->data = dataCopy;
	packetToLunar = kmalloc(sizeof(QUEUED_PACKET), 0);
	if (packetToLunar) {
	  packetToLunar->buffer = skb;
	  packetToLunar->length = sizeof(struct sk_buff); // Not used
	  packetToLunar->direction = NDIS_DIR_RECEIVE;    // Not used
	  NdisInitializeListHead(&packetToLunar->listEntry);
	  NdisInterlockedInsertTailList(&l2rcvQueue, &packetToLunar->listEntry, &l2rcvLock);
	  /* Indicate that a handler thread should take
	     the packet */
	  activate_packet_pump(pAdapt);
	}
      }
    }

    /* Indicate that the packet can be dropped (in order
       to reduce outgoing network traffic) */
    return TRUE;
  }

  /* Drop everything else when Lunar is active */
  return TRUE;
}


/* Returns TRUE if the given packet was "consumed"
   by this function and does not need to be handled
   further */
BOOLEAN
AnalyzePacket(
	      IN PNDIS_PACKET ThePacket,
	      IN COMDIRECTION Direction,
	      IN PADAPT pAdapt)
{
  UINT PhysicalBufferCount = 0;
  UINT BufferCount = 0;
  PNDIS_BUFFER FirstBuffer = NULL;
  PNDIS_BUFFER NextBuffer = NULL;
  UINT TotalPacketLength = 0;
  PVOID VirtualAddress;
  BYTE* MyData;
  UINT Length = 0;
  UINT i, BufIndex;
  BOOLEAN retval;

  // Print some info from the packet
  NdisQueryPacket(ThePacket, &PhysicalBufferCount, &BufferCount,
		  &FirstBuffer, &TotalPacketLength);

  // Disregard from packets that cannot be IP, ARP or
  // SAPF; Very harsh way to distinguish between
  // different packet types
  // TODO: Perhaps restore later
  //if (TotalPacketLength < 32) {
  /* Indicate for the packet to be handled
     elsewhere */
  //return FALSE;
  //}

  // Loop trough the buffers in order to print them all
  NextBuffer = FirstBuffer;

  /* If the given data is under a specific length, then we
     can use our statically allocated packet data buffer
     (in order to save memory allocation cycles) */
  if (TotalPacketLength <= STATIC_DATA_SIZE) {
    MyData = MyStaticData;
  } else {
    MyData = (BYTE*)kmalloc(TotalPacketLength, 0);
    if (!MyData) {
      DbgPrint("--> Passthru - Could not allocate memory for packet <--");
      /* Indicate packet not handled here */
      return FALSE;
    }
  }

  /* Merge the buffers into one long data array */
  i = 0;
  for (BufIndex = 0; BufIndex < BufferCount; BufIndex++) {
    /* Get the next buffer from the packet */
    NdisQueryBufferSafe(NextBuffer,
			&VirtualAddress,
			&Length,
			16 /*NormalPagePriority*/);
    memcpy(MyData + i, (BYTE*)VirtualAddress, Length);
    i += Length;

    // Get the next buffer in the chain, if there is one
    NdisGetNextBuffer(NextBuffer, &NextBuffer);
  }

  if (MyData == MyStaticData) {
    retval = AnalyzePacketBuffer(MyData, TotalPacketLength, Direction, TRUE, pAdapt);
  } else {
    retval = AnalyzePacketBuffer(MyData, TotalPacketLength, Direction, FALSE, pAdapt);
    /* Note: We indicate that the given data is not
       allocated to "volatile" memory so that it
       will have to be released by the last "user";
       Therefore no kfree(MyData) is needed here */
  }

  return retval;
}
